GRADIENT = 'gradient'
CATEGORICAL = 'categorical'

color_style_dict = {
    'd3.schemeCategory10': {'style': CATEGORICAL, 'length': 10},
    'd3.schemeAccent': {'style': CATEGORICAL, 'length': 8},
    'd3.schemeDark2': {'style': CATEGORICAL, 'length': 8},
    'd3.schemePaired': {'style': CATEGORICAL, 'length': 12},
    'd3.schemePastel1': {'style': CATEGORICAL, 'length': 9},
    'd3.schemePastel2': {'style': CATEGORICAL, 'length': 8},
    'd3.schemeSet1': {'style': CATEGORICAL, 'length': 9},
    'd3.schemeSet2': {'style': CATEGORICAL, 'length': 8},
    'd3.schemeSet3': {'style': CATEGORICAL, 'length': 12},
    'd3.schemeTableau10': {'style': CATEGORICAL, 'length': 10},
    'd3.interpolateBrBG': {'style': GRADIENT},
    'd3.interpolatePRGn': {'style': GRADIENT},
    'd3.interpolatePiYG': {'style': GRADIENT},
    'd3.interpolatePuOr': {'style': GRADIENT},
    'd3.interpolateRdBu': {'style': GRADIENT},
    'd3.interpolateRdGy': {'style': GRADIENT},
    'd3.interpolateRdYlBu': {'style': GRADIENT},
    'd3.interpolateRdYlGn': {'style': GRADIENT},
    'd3.interpolateSpectral': {'style': GRADIENT},
    'd3.interpolateBlues': {'style': GRADIENT},
    'd3.interpolateGreens': {'style': GRADIENT},
    'd3.interpolateOranges': {'style': GRADIENT},
    'd3.interpolatePurples': {'style': GRADIENT},
    'd3.interpolateReds': {'style': GRADIENT},
    'd3.interpolateGreys': {'style': GRADIENT},
    'd3.interpolateTurbo': {'style': GRADIENT},
    'd3.interpolateViridis': {'style': GRADIENT},
    'd3.interpolateInferno': {'style': GRADIENT},
    'd3.interpolateMagma': {'style': GRADIENT},
    'd3.interpolatePlasma': {'style': GRADIENT},
    'd3.interpolateCividis': {'style': GRADIENT},
    'd3.interpolateWarm': {'style': GRADIENT},
    'd3.interpolateCool': {'style': GRADIENT},
    'd3.interpolateCubehelixDefault': {'style': GRADIENT},
    'd3.interpolateBuGn': {'style': GRADIENT},
    'd3.interpolateBuPu': {'style': GRADIENT},
    'd3.interpolateGnBu': {'style': GRADIENT},
    'd3.interpolateOrRd': {'style': GRADIENT},
    'd3.interpolatePuBuGn': {'style': GRADIENT},
    'd3.interpolatePuBu': {'style': GRADIENT},
    'd3.interpolatePuRd': {'style': GRADIENT},
    'd3.interpolateRdPu': {'style': GRADIENT},
    'd3.interpolateYlGnBu': {'style': GRADIENT},
    'd3.interpolateYlGn': {'style': GRADIENT},
    'd3.interpolateYlOrBr': {'style': GRADIENT},
    'd3.interpolateYlOrRd': {'style': GRADIENT},
    'd3.interpolateRainbow': {'style': GRADIENT}
}
