files_config = {
    "alias_de": "aliases.de.tsv.gz",
    "alias": "aliases.en.tsv.gz",
    "alias_es": "aliases.es.tsv.gz",
    "alias_fr": "aliases.fr.tsv.gz",
    "alias_it": "aliases.it.tsv.gz",
    "alias_nl": "aliases.nl.tsv.gz",
    "alias_pl": "aliases.pl.tsv.gz",
    "alias_pt": "aliases.pt.tsv.gz",
    "alias_ru": "aliases.ru.tsv.gz",
    "alias_sv": "aliases.sv.tsv.gz",
    "alias_all": "aliases.tsv.gz",
    "alias_zhcn": "aliases.zh-cn.tsv.gz",
    "all": "all.tsv.gz",
    "commondsmedia": "claims.commonsMedia.tsv.gz",
    "external_id": "claims.external-id.tsv.gz",
    "geo_shape": "claims.geo-shape.tsv.gz",
    "globe_coordinate": "claims.globe-coordinate.tsv.gz",
    "math": "claims.math.tsv.gz",
    "monolingualtext": "claims.monolingualtext.tsv.gz",
    "musical_notation": "claims.musical-notation.tsv.gz",
    "claims_other": "claims.other.tsv.gz",
    "quantity": "claims.quantity.tsv.gz",
    "string": "claims.string.tsv.gz",
    "tabular_data": "claims.tabular-data.tsv.gz",
    "time": "claims.time.tsv.gz",
    "claims": "claims.tsv.gz",
    "url": "claims.url.tsv.gz",
    "wikibase_form": "claims.wikibase-form.tsv.gz",
    "item": "claims.wikibase-item.tsv.gz",
    "wikibase_lexeme": "claims.wikibase-lexeme.tsv.gz",
    "wikibase_property": "claims.wikibase-property.tsv.gz",
    "wikibase_sense": "claims.wikibase-sense.tsv.gz",
    "dwd_isa": "derived.dwd_isa.tsv.gz",
    "isastar": "derived.isastar.tsv.gz",
    "isa": "derived.isa.tsv.gz",
    "p279star": "derived.P279star.tsv.gz",
    "p279": "derived.P279.tsv.gz",
    "p31p279star": "derived.P31P279star.tsv.gz",
    "p31": "derived.P31.tsv.gz",
    "description_de": "descriptions.de.tsv.gz",
    "description": "descriptions.en.tsv.gz",
    "description_es": "descriptions.es.tsv.gz",
    "description_fr": "descriptions.fr.tsv.gz",
    "description_it": "descriptions.it.tsv.gz",
    "description_nl": "descriptions.nl.tsv.gz",
    "description_pl": "descriptions.pl.tsv.gz",
    "description_pt": "descriptions.pt.tsv.gz",
    "description_ru": "descriptions.ru.tsv.gz",
    "description_sv": "descriptions.sv.tsv.gz",
    "description_all": "descriptions.tsv.gz",
    "description_zhcn": "descriptions.zh-cn.tsv.gz",
    "label_de": "labels.de.tsv.gz",
    "label": "labels.en.tsv.gz",
    "KGTK_LABEL_FILE": "labels.en.tsv.gz",
    "label_es": "labels.es.tsv.gz",
    "label_fr": "labels.fr.tsv.gz",
    "label_it": "labels.it.tsv.gz",
    "label_nl": "labels.nl.tsv.gz",
    "label_pl": "labels.pl.tsv.gz",
    "label_pt": "labels.pt.tsv.gz",
    "label_ru": "labels.ru.tsv.gz",
    "label_sv": "labels.sv.tsv.gz",
    "label_all": "labels.tsv.gz",
    "label_zhcn": "labels.zh-cn.tsv.gz",
    "in_degree": "metadata.in_degree.tsv.gz",
    "out_degree": "metadata.out_degree.tsv.gz",
    "pagerank_directed": "metadata.pagerank.directed.tsv.gz",
    "pagerank_undirected": "metadata.pagerank.undirected.tsv.gz",
    "datatypes_augmented": "metadata.property.datatypes.augmented.tsv.gz",
    "datatypes": "metadata.property.datatypes.tsv.gz",
    "types": "metadata.types.tsv.gz",
    "qualifiers_commonsmedia": "qualifiers.commonsMedia.tsv.gz",
    "qualifiers_external_id": "qualifiers.external-id.tsv.gz",
    "qualifiers_geo_shape": "qualifiers.geo-shape.tsv.gz",
    "qualifiers_globe_coordinate": "qualifiers.globe-coordinate.tsv.gz",
    "qualifiers_math": "qualifiers.math.tsv.gz",
    "qualifiers_monolingual_text": "qualifiers.monolingualtext.tsv.gz",
    "qualifiers_musical_notation": "qualifiers.musical-notation.tsv.gz",
    "qualifiers_quantity": "qualifiers.quantity.tsv.gz",
    "qualifiers_string": "qualifiers.string.tsv.gz",
    "qualifiers_tabular_data": "qualifiers.tabular-data.tsv.gz",
    "qualifiers_time": "qualifiers.time.tsv.gz",
    "qualifiers": "qualifiers.tsv.gz",
    "qualifiers_url": "qualifiers.url.tsv.gz",
    "qualifiers_wikibase_form": "qualifiers.wikibase-form.tsv.gz",
    "qualifiers_wikibase_item": "qualifiers.wikibase-item.tsv.gz",
    "qualifiers_wikibase_lexeme": "qualifiers.wikibase-lexeme.tsv.gz",
    "qualifiers_wikibase_property": "qualifiers.wikibase-property.tsv.gz",
    "qualifiers_wikibase_sense": "qualifiers.wikibase-sense.tsv.gz",
    "sitelinks_en": "sitelinks.en.tsv.gz",
    "sitelinks_qualifiers_en": "sitelinks.qualifiers.en.tsv.gz",
    "sitelinks_qualifiers": "sitelinks.qualifiers.tsv.gz",
    "sitelinks": "sitelinks.tsv.gz"
}
